# -*- coding: UTF-8 -*-
__author__ = "d01"
__copyright__ = "Copyright (C) 2015, Florian JUNG"
__license__ = "All rights reserved"
__version__ = "0.1.0a0"
__date__ = "2015-03-13"
# Created: 2015-03-13 12:11
